/***************************************************************************
 *   Copyright (C) 2010 by TAM(Teppei Tamra)                               *
 *   tam-t@par.odn.ne.jp                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "ivglview.h"
#include <math.h>
#include <QDebug>
#include <QTimer>

#define POINT(x,y) (x)*(density+1)+(y)

IVGLView::IVGLView(QString &filename,QWidget *parent) :
    QGLWidget(parent)
{
    angle = 0;
    posX = 0;
    posY = 0;
    zoom = 1;
    line = false;
    image = true;
    weight = false;
    movelock = true;
    lbutton = false;
    rbutton = false;
    ppoint = QPoint(0,0);
    density = 80;
    motionCount = 0;
    load(filename);
    //load(":/images/f.png");
}

bool IVGLView::load(const QString &filename)
{
    QImage ti;
    if (!ti.load(filename)) return false;
    img = ti.scaled(ti.width() / 2,ti.height(),Qt::IgnoreAspectRatio,Qt::SmoothTransformation);
    tex = QGLWidget::convertToGLFormat(img);
    QImage simg;
    QImage wimg;
    if (!wimg.load(".ivw_"+filename)) wimg = img;
    simg = wimg.scaled(density,density,Qt::IgnoreAspectRatio,Qt::SmoothTransformation);
    for(int i = 0;i <= density;i ++) {
        for(int j = 0;j <= density;j ++) {
            IVGLViewPoint p;
            p.angle = 0;
            p.originX = i * (1 / density);
            if (p.originX > 1.0) p.originX = 1.0;
            p.originY = j * (1 / density);
            if (p.originY >1.0) p.originY = 1.0;
            p.nowX = p.originX;
            p.nowY = p.originY;
            // [p1][p2]
            // [p3][p4]
            QColor p1 = simg.pixel(i - 1,density - 1 - j);
            QColor p2 = simg.pixel(i,density - 1 - j);
            QColor p3 = simg.pixel(i - 1,density - 1 - j + 1);
            QColor p4 = simg.pixel(i,density - 1 - j + 1);
            p.weight = (double)(qGray(p1.rgb()) + qGray(p2.rgb()) + qGray(p3.rgb()) + qGray(p4.rgb())) / (4 * 256);


            points.append(p);
        }
    }

    glBindTexture(GL_TEXTURE_2D,texture[0]);
    emit process();
    return true;
}


void IVGLView::initializeGL()
{
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_MULTISAMPLE);

//    glClearColor(0.9, 0.9, 0.9, 1.0);
//    glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);

    glEnable(GL_TEXTURE_2D);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glGenTextures(1,&texture[0]);
    glBindTexture(GL_TEXTURE_2D,texture[0]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    //glTexImage2D(GL_TEXTURE_2D, 0, 3, tex.width(), tex.height(), 0, GL_RGBA,GL_UNSIGNED_BYTE, tex.bits());
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, tex.width(), tex.height(), GL_RGBA,GL_UNSIGNED_BYTE, tex.bits());

//    glEnable(GL_LIGHTING);
//    glEnable(GL_LIGHT0);
//    static GLfloat lightPosition[4] = { 0.5, 5.0, 7.0, 1.0 };
//    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

}

void IVGLView::resizeGL(int width, int height)
{
    double as = ((double)height / (double)width);// / 2.0;
    glViewport(0, 0, width, height);

//    gluPerspective(30,width / height,0.001,1.0);
////    glOrtho(-1.0, 1.0, 0.0, 1.0, -1.0, 1.0);
//    glFrustum(-1.0, 1.0, -1.0*h, 1.0*h,0, 1.0);
    glMatrixMode(GL_PROJECTION);
    glOrtho(-1.0, 1.0, -1.0*as, 1.0*as, -1.0, 1.0);
//    glEnable(GL_TEXTURE_2D);
}

void IVGLView::paintGL()
{
    double h = ((double)tex.height() / (double)tex.width());// / 2.0;
    double as = ((double)this->height() / (double)this->width());
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0*as, 1.0*as, -1.0, 1.0);
    //glFrustum(-1.0, 1.0, -1.0*h, 1.0*h,-1, 1.0);


    glPushMatrix();
    glTranslated((2.0 / (double)width()) * (double)posX,
                 (2.0 / (double)height()) * (double)posY * h,
                 0);
    glRotated(angle,0,0,1);
    glScaled(zoom,zoom,0);




    for(int i = 0;i < density;i ++) {
        for(int j = 0;j < density;j ++) {
/*
            IVGLViewPoint p1 = points.at(((i)*(density + 1))+(j));
            IVGLViewPoint p2 = points.at(((i+1)*(density + 1))+(j));
            IVGLViewPoint p3 = points.at(((i+1)*(density + 1))+(j+1));
            IVGLViewPoint p4 = points.at(((i)*(density + 1))+(j+1));
*/
            IVGLViewPoint p1 = points.at(POINT(i,j));
            IVGLViewPoint p2 = points.at(POINT(i+1,j));
            IVGLViewPoint p3 = points.at(POINT(i+1,j+1));
            IVGLViewPoint p4 = points.at(POINT(i,j+1));

            if (image) {
                glColor4f(1.0,1.0,1.0,0.5);
//            glColor3f(p1.weight,p1.weight,p1.weight);
                glBegin(GL_QUADS);
                glTexCoord3d(p1.originX,p1.originY,0);
                glVertex3d(-1 + (p1.nowX*2),(-1 + (p1.nowY*2))*h,0);
                glTexCoord3d(p2.originX,p2.originY,0);
                glVertex3d(-1 + (p2.nowX*2),(-1 + (p2.nowY*2))*h,0);
                glTexCoord3d(p3.originX,p3.originY,0);
                glVertex3d(-1 + (p3.nowX*2),(-1 + (p3.nowY*2))*h,0);
                glTexCoord3d(p4.originX,p4.originY,0);
                glVertex3d(-1 + (p4.nowX*2),(-1 + (p4.nowY*2))*h,0);
                glEnd();
            }

            if (weight) {
                glBegin(GL_QUADS);
                glColor3f(p1.weight,p1.weight,p1.weight);
                glVertex2d(-1 + (p1.nowX*2),(-1 + (p1.nowY*2))*h);
                glColor3f(p1.weight,p2.weight,p2.weight);
                glVertex2d(-1 + (p2.nowX*2),(-1 + (p2.nowY*2))*h);
                glColor3f(p1.weight,p3.weight,p3.weight);
                glVertex2d(-1 + (p3.nowX*2),(-1 + (p3.nowY*2))*h);
                glColor3f(p1.weight,p4.weight,p4.weight);
                glVertex2d(-1 + (p4.nowX*2),(-1 + (p4.nowY*2))*h);
                glEnd();
            }

            if (line) {
                glColor3f(1.0,0.0,0.0);
//            glColor4f(p1.weight,p1.weight,p1.weight,0.5);
                glBegin(GL_LINE_LOOP);
                glVertex3d(-1 + (p1.nowX*2),(-1 + (p1.nowY*2))*h,0);
                glVertex3d(-1 + (p2.nowX*2),(-1 + (p2.nowY*2))*h,0);
                glVertex3d(-1 + (p3.nowX*2),(-1 + (p3.nowY*2))*h,0);
                glVertex3d(-1 + (p4.nowX*2),(-1 + (p4.nowY*2))*h,0);
                glEnd();
            }
        }
    }


    glPopMatrix();

//    glMatrixMode(GL_MODELVIEW);
//    glLoadIdentity();
//    glRotated(20,20,20,1);
//    glMatrixMode(GL_PROJECTION);


    glFlush();
//    glDrawPixels(img.width(),img.height(),GL_RGBA,GL_UNSIGNED_BYTE,tex.bits());
}

void IVGLView::mousePressEvent(QMouseEvent *event)
{
    ppoint = event->pos();
    if (event->button() == Qt::LeftButton) lbutton = true;
    if (event->button() == Qt::RightButton) rbutton = true;
    return;
}

void IVGLView::mouseReleaseEvent(QMouseEvent *event)
{
    ppoint = event->pos();
    if (event->button() == Qt::LeftButton) lbutton = false;
    if (event->button() == Qt::RightButton) rbutton = false;
}

void IVGLView::mouseMoveEvent(QMouseEvent *event)
{
    if ((!rbutton) && (!lbutton)) {
        ppoint = event->pos();
        return;
    }
    if (lbutton) {
        if (!movelock) {
            posX += (event->x() - ppoint.x());
            posY += (event->y() - ppoint.y()) * -1;
        }
        motion(event->x() - ppoint.x(),(event->y() - ppoint.y()) * -1 );
        update();
    }
//    if (rbutton) {
//        angle += (event->x() - ppoint.x());
//        zoom += (double)(event->y() - ppoint.y()) / 20.0;
//        update();
//    }
    ppoint = event->pos();
    return;
}

void IVGLView::shift(int x)
{
    motion(x,0);
}

void IVGLView::wheelEvent(QWheelEvent *event)
{
    if (rbutton) {
        angle += (double)(event->delta()) / 45;
    } else {
        zoom += 4.0 / (double)(event->delta());
    }
    update();
    return;
}

void IVGLView::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_L) {
        if (line) line = false;
        else line = true;
    } else if (event->key() == Qt::Key_I) {
        if (image) image = false;
        else image = true;
    } else if (event->key() == Qt::Key_W) {
        if (weight) weight = false;
        else weight = true;
    } else if (event->key() == Qt::Key_Period) {
        if (movelock) movelock = false;
        else movelock = true;
    }
    update();
    return;
}

void IVGLView::process()
{
//    qDebug() << ".";
    if (motionCount) {
        counterAction();
        update();
    }
    //QTimer::singleShot(33,this,SLOT(process()));
    return;
}

void IVGLView::motion(qreal x,qreal y)
{
    double mx = (double)x / (double)img.width();
    double my = (double)y / (double)img.height();

    for(int i = 0;i < points.count();i ++) {
        if ((i > density) &&
                (i < points.count() - density) &&
                (i % ((int)density + 1) != 0) &&
                (i % ((int)density + 1) != density)) {
            IVGLViewPoint p = points.at(i);
            p.nowX += (mx * (p.weight));
            p.nowY += (my * (p.weight));
            points[i] = p;
        }
    }

    if (mx < 0) {
        for (int i = 1;i < density;i ++) {
            for (int j = 1;j < density;j ++) {
                IVGLViewPoint p = points.at(POINT(i,j));
                if (points.at(POINT(i - 1,j)).nowX > p.nowX) points[POINT(i,j)].nowX = points.at(POINT(i - 1,j)).nowX;
            }
        }
    } else {
        for (int i = density - 1;i > 0;i --) {
            for (int j = density - 1;j > 0;j --) {
                IVGLViewPoint p = points.at(POINT(i,j));
                if (points.at(POINT(i + 1,j)).nowX < p.nowX) points[POINT(i,j)].nowX = points.at(POINT(i + 1,j)).nowX;
            }
        }
    }
    if (my < 0) {
        for (int i = 1;i < density;i ++) {
            for (int j = 1;j < density;j ++) {
                IVGLViewPoint p = points.at(POINT(i,j));
                if (points.at(POINT(i,j - 1)).nowY > p.nowY) points[POINT(i,j)].nowY = points.at(POINT(i,j - 1)).nowY;
            }
        }
    } else {
        for (int i = density - 1;i > 0;i --) {
            for (int j = density - 1;j > 0;j --) {
                IVGLViewPoint p = points.at(POINT(i,j));
                if (points.at(POINT(i,j + 1)).nowY < p.nowY) points[POINT(i,j)].nowY = points.at(POINT(i,j + 1)).nowY;
            }
        }
    }

    /*
    for (int i = 1;i < density - 1;i ++) {
        for (int j = 1;j < density - 1;j ++) {
            IVGLViewPoint p = points.at(POINT(i,j));
            if (mx > 0) {
                if (i < density) {
                    if (points.at(POINT(i + 1,j)).nowX < p.nowX) p.nowX = points.at(POINT(i + 1,j)).nowX;
                }
            } else {
                if (i > 0) {
                    if (points.at(POINT(i - 1,j)).nowX > p.nowX) p.nowX = points.at(POINT(i - 1,j)).nowX;
                }
            }
            if (my > 0) {
                if (j < density) {
                    if (points.at(POINT(i,j + 1)).nowY < p.nowY) p.nowY = points.at(POINT(i,j + 1)).nowY;
                }
            } else {
                if (j > 0) {
                    if (points.at(POINT(i,j - 1)).nowY > p.nowY) p.nowY = points.at(POINT(i,j - 1)).nowY;
                }
            }
            points[POINT(i,j)] = p;
        }
    }
*/
    motionCount = 60;
    return;
}

void IVGLView::counterAction()
{
    for(int i = 0;i < points.count();i ++) {
        IVGLViewPoint p = points.at(i);
        if ((p.nowX != p.originX) || (p.nowY != p.originY)) {
            if (p.nowX > p.originX) p.nowX = p.nowX - ((p.nowX - p.originX) / 10);
            else p.nowX = p.nowX + ((p.originX - p.nowX) / 10);
            if (p.nowY > p.originY) p.nowY = p.nowY - ((p.nowY - p.originY) / 10);
            else p.nowY = p.nowY + ((p.originY - p.nowY) / 10);
            points[i] = p;
        }
    }
    motionCount --;
    return;
}
