/***************************************************************************
 *   Copyright (C) 2010 by TAM(Teppei Tamra)                               *
 *   tam-t@par.odn.ne.jp                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IVGLVIEW_H
#define IVGLVIEW_H

#include <QGLWidget>
#include <QImage>
#include <QMouseEvent>
#include <QWheelEvent>
#include <QList>
#include "ivimage.h"


class IVGLViewPoint
{
public:
    double originX;
    double originY;
    double nowX;
    double nowY;
    double weight;
    double angle;
    double power;
};

class IVGLView : public QGLWidget
{
    Q_OBJECT
public:
    explicit IVGLView(QString &filename,QWidget *parent = 0);
    bool load(const QString &filename);
    void shift(int x);

protected:
    virtual void initializeGL();
    virtual void paintGL();
    virtual void resizeGL(int width,int height);
    virtual void mousePressEvent(QMouseEvent *event);
    virtual void mouseReleaseEvent(QMouseEvent *event);
    virtual void mouseMoveEvent(QMouseEvent *event);
    virtual void wheelEvent(QWheelEvent *event);
    virtual void keyPressEvent(QKeyEvent * event);
    void motion(qreal x,qreal y);
    void counterAction();

protected slots:
    void process();

signals:

protected:
    QImage tex;
    QImage img;
    GLuint texture[256];
    qreal angle;
    qreal zoom;
    qreal posX;
    qreal posY;
    bool lbutton;
    bool rbutton;
    QPoint ppoint;
    QList<IVGLViewPoint> points;
    double density;
    bool image;
    bool line;
    bool weight;
    bool movelock;
    int motionCount;
public slots:

};

#endif // IVGLVIEW_H
